/*
 * Copyright (C) 1996 Michael Davidson.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#pragma ident	"@(#)mmap.c	1.7	99/02/18 SMI"

#include	<errno.h>
#include	<stdio.h>
#include	<fcntl.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<sys/mman.h>
#include	<sys/param.h>
#include	"lxrun.h"
#include	"linux/types.h"
#include	"linux/mman.h"

/*
 * mmap() related system calls
 */

/*
 *	mmap
 *	munmap
 *	mprotect
 *	msync		(NOT IMPLEMENTED)
 *	mlock		(NOT IMPLEMENTED)
 *	munlock		(NOT IMPLEMENTED)
 *	mlockall	(NOT IMPLEMENTED)
 *	munlockall	(NOT IMPLEMENTED)
 */

int dzfd = -1;

void *
dz_mmap(void *addr, size_t len, int prot, int flags, off_t off)
{
	if (dzfd == -1) {
		dzfd = open("/dev/zero", O_RDWR);
		fcntl(dzfd, F_SETFD, FD_CLOEXEC);
	}
	if (dzfd == -1)
		return MAP_FAILED;

	return mmap(addr, len, prot, flags, dzfd, off);
}

void *
lx_mmap(unsigned long *av)
{
	void	*addr;
	size_t	len;
	int	lx_prot;
	int	lx_flags;
	int	fd;
	off_t	off;
	int	prot	= 0;
	int	flags	= 0;

	if (! verify_read(av, 6 * sizeof(av[0])))
		return (void *)-1;

	addr	= (void *)av[0];
	len	= (size_t)av[1];
	lx_prot	= (int)av[2];
	lx_flags= (int)av[3];
	fd	= (int)av[4];
	off	= (off_t)av[5];

	if (lx_prot & LX_PROT_READ)
		prot |= PROT_READ;
	if (lx_prot & LX_PROT_WRITE)
		prot |= PROT_WRITE;
	if (lx_prot & LX_PROT_EXEC)
		prot |= PROT_EXEC;

	if (lx_flags & LX_MAP_SHARED)
		flags |= MAP_SHARED;
	if (lx_flags & LX_MAP_PRIVATE)
		flags |= MAP_PRIVATE;
	if (lx_flags & LX_MAP_FIXED)
		flags |= MAP_FIXED;

	if (lx_flags & LX_MAP_ANONYMOUS)
		return dz_mmap(addr, len, prot, flags, off);

	return mmap(addr, len, prot, flags, fd, off);
}	

int
lx_munmap(void *addr, size_t len)
{
	return munmap(addr, len);
}	

int
lx_mprotect(void *addr, size_t len, int lx_prot)
{
	int	prot	= 0;

	if (lx_prot & LX_PROT_READ)
		prot |= PROT_READ;
	if (lx_prot & LX_PROT_WRITE)
		prot |= PROT_WRITE;
	if (lx_prot & LX_PROT_EXEC)
		prot |= PROT_EXEC;

#ifdef OSR5
        if(mprotect(addr, len, (prot==PROT_NONE)
                                        ? PROT_READ : PROT_NONE) < 0 &&
                                        errno != ENOMEM)
                return -1;
#endif

	return mprotect(addr, len, prot);
}

int lx_msync()		{ return enosys("msync"); }
int lx_mlock()		{ return enosys("mlock"); }
int lx_munlock()	{ return enosys("munlock"); }
int lx_mlockall()	{ return enosys("mlockall"); }
int lx_munlockall()	{ return enosys("munlockall"); }

void *
lx_mremap(void *oaddr, size_t old_size, size_t new_size, unsigned long flags)
{
	if (new_size < old_size) {
		uintptr_t oend = (uintptr_t)oaddr + PAGE_ALIGN(old_size);
		uintptr_t nend = (uintptr_t)oaddr + PAGE_ALIGN(new_size);
		if (munmap((void *)nend, oend - nend) == -1) {
			perror("munmap");
		}
		return oaddr;
	} else {
		tprintf("can't lx_mremap"
		    "(oaddr = %p, osize = %x, nsize = %x, flags = %x)\n",
		    oaddr, old_size, new_size, flags);
		return (void *)enosys("mremap(map growing)");
	}
}
